-- Document
--
-- ICFP 2001 programming contest
-- Tom Moertel <tom-icfp2001@moertel.com>

module Document where

newtype Document = Document [DocPart] deriving (Eq, Ord)

data DocPart = DText String
             | DB Document
             | DEM Document
             | DI Document
             | DPL Document
             | DS Document
             | DTT Document
             | DU Document
             | DSize Int Document -- 0 thru 9 are legal (we use -1 as special)
             | DColor Color Document
             deriving (Eq, Ord)

data Color =
    Red | Green | Blue | Cyan | Magenta | Yellow | Black | White | AnyColor
    deriving (Show, Eq, Ord, Enum, Bounded)

instance Show Document where
    showsPrec _ = showsDoc

instance Show DocPart where
    showsPrec _ = showsDP

showsDP :: DocPart -> String -> String
showsDP (DText s)          	=  (s ++)
showsDP (DB d)             	=  showsTag "B"d
showsDP (DEM d)            	=  showsTag "EM"d
showsDP (DI d)             	=  showsTag "I"d
showsDP (DPL d)            	=  showsTag "PL"d
showsDP (DS d)             	=  showsTag "S"d
showsDP (DTT d)            	=  showsTag "TT"d
showsDP (DU d)             	=  showsTag "U"d
showsDP (DSize s d)        	=  (showsTag $! (show s)) d
showsDP (DColor Red d)     	=  showsTag "r"d
showsDP (DColor Green d)   	=  showsTag "g"d
showsDP (DColor Blue d)    	=  showsTag "b"d
showsDP (DColor Cyan d)    	=  showsTag "c"d
showsDP (DColor Magenta d) 	=  showsTag "m"d
showsDP (DColor Yellow d)  	=  showsTag "y"d
showsDP (DColor Black d)   	=  showsTag "k"d
showsDP (DColor White d)   	=  showsTag "w"d
showsDP (DColor AnyColor d)	=  showsTag "ANYCOLOR"d -- special

showsTag :: String -> Document -> String -> String
showsTag tag d =
    (('<':tag) ++) .  ('>':) . showsDoc d . (('<':'/':tag) ++) . ('>':)

showDoc :: Document -> String
showDoc d = showsDoc d ""

showsDoc :: Document -> String -> String
showsDoc (Document dparts) s =
    foldr showsDP s dparts
