-- NaiveOptimization - a very simple 1st pass optimization
--
-- ICFP 2001 programming contest
-- Tom Moertel <tom-icfp2001@moertel.com>

-- The idea is to quickly get the document stream into a canonical
-- form from which to start the serious optimizations.  So, we
-- take the original document, convert it into its ``meaning'' and
-- then translate the meaning back into a document by looking for
-- changes in its decorated ouput.  If the back-translated document
-- is smaller than the original, we keep it; otherwise, we throw
-- it out and return the original.

module NaiveOptimization (naiveOptimizer,optDiff) where

import IOExts
import IO

import Document
import Meaning
import PropRecUtils

naiveOptimizer :: Document -> Document
naiveOptimizer doc =
    if length (show doc') <= length (show doc)
    then doc'
    else unsafePerformIO (logMsg ("OPTIMIZED " ++ (show doc') ++ " IS LARGER") >> return doc)
    where meaning = docMeaning doc
          doc' = Document $ fst $ m2d [rootProps] meaning

m2d :: [PropRec] -> Meaning -> ([DocPart], Meaning)
m2d _ []     = ([], [])
m2d [] _     = error "property record stack underflow in naive optimizer"
m2d rstack@(r:_) mstream@(m@(mr,_):_)
    | matchQ r m = let (span, mstream')  = textSpan r mstream
                       (doc', mstream'') = m2d rstack mstream'
                   in (DText span : doc', mstream'')
    | otherwise  = let diffPREs = [ pre | pre <- propTable
                                  , propCompare pre r mr == False ]
                       builders = [ propUpBuilder pre r mr | pre  <- diffPREs ]
                       r's      = [ propUp pre mr r | pre <- diffPREs ]
                   in  case [ x | x@Nothing <- builders ]  of
                           [] -> -- we can build up to the desired decoration
                                 -- TODO: tk best
                                 let Just docCtor  = head builders
                                     r'            = head r's
                                     (doc', ms')   = m2d (r':rstack) mstream
                                     (doc'', ms'') = m2d rstack ms'
                                 in  (docCtor (Document doc') : doc'', ms'')
                           _  -> -- we can't build up, so we must back down
                                 ([], mstream)

textSpan :: PropRec -> Meaning -> (String, Meaning)
textSpan _ []     = ([], [])
textSpan r (m:ms) =
    if matchQ r m
    then let (s', ms') = textSpan r ms in (snd m : s', ms')
    else ([], m:ms)
    
optDiff :: Document -> [String]
optDiff doc = meaningDiff doc (naiveOptimizer doc)

logMsg :: String -> IO ()
logMsg s = hPutStrLn stderr s
